' Integrated 1st-person game actor.

' Implementing a classic 3d game actor by conveniently mixing physics And
' old-style 3d game coding.

' You can easily implement a 1st-person-view game by just placing a camera
' inside a physics-enabled ball And push it around, but If you need To
' perform classic, non-physics behavior For your actor, this sample code
' may be a solid basis For your work.

' The technique:
' - a simple vector represents the current 1st-person actor location. The actor
'   is moved around by changing the vector, like old-style 3d games.
'   Sloping terrain is detected by scanning (iBodyScan()) the scenery body And
'   horizontal actor movements are opposed proportionally To the steepness.
' - a physics enabled ball is 'linked' to the changing actor-location by a spring.
'   When the actor location vector changes, the spring pulls the ball toward it.
' - the camera is attached To the ball center. The view point is from inside the ball.
'   But camera orientation doesn't follow ball orientation. Instead, it is mouse controlled.
' - the ball surrounding the camera (actor shell) will Not go through scenery features And
'   solid objects, as it is a solid body too. So the viewpoint will never go through things.
'   Actor shell collisions are monitored And the actor location vector is adjusted
'   accordingly To perform proper collision response.

' THIS SOURCE CODE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND. USE AT YOUR OWN RISK.

Strict
Framework _3Impact.Lib

Global Camera:CAMERA
Global FontWhite:FONT
Global LoadingMessage:SPRITE
Global SkyBox:SKYBOX
Global Terrain:BODY
Global TerrainMesh:MESH
Global Ball:BODY
Global BallMesh:MESH
Global BallMeshShadow:MESH
Global FrustumMesh:MESH
Global FrustumMeshShadow:MESH
Global ActorShell:BODY
Global ActorShellMesh:MESH
Global ActorShellMeshShadow:MESH
Global BallTerrain:BODYBODY
Global ActorShellTerrain:BODYBODY
Global ActorShellBall:BODYBODY
Global Spring:SPRING
Global ActorLocation:D3DXVECTOR3=_D3DXVECTOR3()
Global ActorVerticalSpeed:Float 
Global MouseLookYRot:Float 
Global MouseLookXRot:Float 
Global DebugViewMode:Int

Const SHELL_RADIUS:Float = 1.0

If Not i3ImpactOpen() RuntimeError "Unable to initialize the 3Impact game engine"
If Not i3ImpactSettings(_ISETTINGS(640,480,32,False,True,False)) RuntimeError "Settings failed"

Local engine:sample_3IEngine=New sample_3IEngine

engine.Start()

Type sample_3IEngine Extends _3IEngine

	Method _Init()
	   Camera=iCameraCreate(0.0,0.0,1.0,1.0)
	   iCameraLocationSet(Camera,_D3DXVECTOR3(10.0,3.0,-5.0))
	
	   FontWhite=iFontCreate("default_res\\fontwhite\\font.wid",Null)
	
	   iLightDirectionalSet(_D3DXVECTOR3(1.0,-1.0,1.0),_D3DXVECTOR4(1.0,1.0,1.0,0.35))
	
	   LoadingMessage=iSpriteCreate("default_res\\sprites\\loading.x",Null)
	   iSpriteRender(LoadingMessage,False)
	
	   SkyBox=iSkyBoxCreate("default_res\\skyboxes\\skybox01.sky")
	
	   Terrain=iBodyCreate("default_res\\terrain_5.00.ply")
	   TerrainMesh=iBodyMeshCreate("default_res\\terrain.x",Terrain)
	
	   ' Create a standard ball To test collisions with the actor shell
	   Ball=iBodySGCreate("default_res\\ball_.spg",0.13)
	   BallMesh=iBodyMeshCreate("default_res\\ball.x",Ball)
	   BallMeshShadow=iMeshBodyShadowCasterCreate("default_res\\ball_shadow.x",Ball,500.0)
	   iBodyLocationSet(Ball,_D3DXVECTOR3(0.0,5.0,0.0),True)
	   iBodyDampingSet(Ball,0.0,0.999,0.0,0.995)
	
	   ' Initialize the actor location vector 
	   ActorLocation = _D3DXVECTOR3(0.0,3.0,-10.0)
	   ' Initialize actor vertical speed 
	   ActorVerticalSpeed = 0.0
	
	   ' Create a simple mesh To visualize the viewing frustum (debug mode only)
	   FrustumMesh=iMeshCreate("default_res\\frustum.x")
	   FrustumMeshShadow=iMeshShadowCasterCreate("default_res\\frustum.x",50.0)
	
	   ' Create the actor shell ball...
	   ActorShell=iBodySGCreate("default_res\\ball_.spg",0.13)
	   ' ...along with its meshes To visualize it (debug mode only)
	   ActorShellMesh=iBodyMeshCreate("default_res\\ballgreen.x",ActorShell)
	   iMeshRenderModeSet(ActorShellMesh,D3DBLEND_SRCCOLOR,D3DBLEND_INVSRCCOLOR,False,True)
	   ActorShellMeshShadow=iMeshBodyShadowCasterCreate("default_res\\ball_shadow.x",ActorShell,500.0)
	   iBodyLocationSet(ActorShell,ActorLocation,True)
	   iBodyDampingSet(ActorShell,0.0,0.8,0.0,0.999)
	
	   ' Enable collision detection between the standard ball And the scenery
	   BallTerrain=iBodyBodyCreate(Ball,Terrain)
	   iBodyBodyFrictionSet(BallTerrain,10.0)
	
	   ' Enable collision detection between the actor shell And the scenery
	   ActorShellTerrain=iBodyBodyCreate(ActorShell,Terrain)
	   iBodyBodyRestitutionSet(ActorShellTerrain,0.0)
	   iBodyBodyFrictionSet(ActorShellTerrain,0.0)
	
	   ' Enable collision detection between the actor shell And the standard ball
	   ActorShellBall=iBodyBodyCreate(ActorShell,Ball)
	   iBodyBodyFrictionSet(ActorShellBall,0.0)
	   iBodyBodyFeedbackEnable(ActorShellBall,-1,0,True)
	
	   ' Create the spring To link the shell To the actor location vector
	   Spring = iSpringCreate(ActorShell,_D3DXVECTOR3(0.0,0.0,0.0),Null,ActorLocation)
	   ' Set spring parameters. They will affect view point responsiveness To actor movements
	   iSpringSet(Spring,10.0)
	   iSpringDampingSet(Spring,0.0)
	
	   ' Camera orientation is independent And performed about Y And X axes. Initialize it.
	   MouseLookYRot = 0.0
	   MouseLookXRot = 0.0
	
	   ' Initialize the flag determining current demo mode. Start in debug mode.
	   DebugViewMode = True
	
	   ' Enable all bodies
	   iBodyEnable(Terrain)
	   iBodyEnable(Ball)
	   iBodyEnable(ActorShell)
	   iSpringEnable(Spring)
	
	   ' Show all meshes
	   iSpriteHide(LoadingMessage)
	   iSkyBoxShow(SkyBox)
	   iMeshShow(TerrainMesh)
	   iMeshShow(BallMesh)
	   iMeshShow(BallMeshShadow)
	   If (DebugViewMode)
	      iMeshShow(ActorShellMesh)
	      iMeshShow(ActorShellMeshShadow)
	      iMeshShow(FrustumMesh)
	      iMeshShow(FrustumMeshShadow)
	   EndIf
	EndMethod
	
	Method MouseLook(mouselookx:Float Var,mlxspd:Float,mouselooky:Float Var,mlyspd:Float,reset:Int)
	   ' A tricky Function To dynamically alter the contents
	   ' of two Float variables by using the mouse.
	
	   ' Float* mouselookx = pointer To the Float variable To alter by moving the mouse up/down
	   ' Float mlxspd = variation speed factor (up/down)
	   ' Float* mouselooky = pointer To the Float variable To alter by moving the mouse Left/Right
	   ' Float mlyspd = variation speed factor (Left/Right)
	   ' BOOL reset = whether To reset (True) the mouselook system
	
	   Global mouselookxold:Float = 0.0
	   Global mouselookyold:Float = 0.0
	   If (reset)
	      mouselookxold = mouselookx / mlxspd
	      mouselookyold = mouselooky / mlyspd
	      iMouseReset(_D3DXVECTOR2(0.5,0.5))
	   Else
	      mouselookx :/ mlxspd
	      mouselooky :/ mlyspd
	      mouselookx = mouselookxold + (0.5-iMouseY())
	      mouselooky = mouselookyold + (iMouseX()-0.5)
	      If (iFloatAbs(mouselookxold - mouselookx) > 0.45)
	         mouselookxold = mouselookx
	         iMouseReset(_D3DXVECTOR2(iMouseX(),0.5))
	      EndIf
	      If (iFloatAbs(mouselookyold - mouselooky) > 0.45)
	         mouselookyold = mouselooky
	         iMouseReset(_D3DXVECTOR2(0.5,iMouseY()))
	      EndIf
	      mouselookx :* mlxspd
	      mouselooky :* mlyspd
	   EndIf
	EndMethod
	
	Method _Run()
	   ' Poll the mouse And update mouse orientation variables accordingly
	   MouseLook(MouseLookXRot,-25.0,MouseLookYRot,35.0,False)
	
	   Local actorshelllocation:D3DXVECTOR3=_D3DXVECTOR3()
	   ' Retrieve And store current location of actor shell
	   iBodyLocationCM(ActorShell,actorshelllocation)
	
	   If Not DebugViewMode
	      ' The demo is in 1st-person view mode.
	      ' Keep the camera attached To the center of ActorShell body
	      iCameraLocationSet(Camera,actorshelllocation)
	      ' Also set camera orientation To current mouse orientation
	      iCameraEulerOrientationSet(Camera,MouseLookXRot,MouseLookYRot,0.0,"xyz")
	   Else
	      ' The demo is in debug view mode.
	      ' The camera is no longer attached To ActorShell object.
	      ' Orient the camera To take the visible actor shell.
	      iCameraLookAt(Camera,actorshelllocation,0.1)
	      ' Also update location And orientation of the visible viewing frustum Object
	      iMeshLocationSet(FrustumMesh,actorshelllocation)
	      iMeshEulerOrientationSet(FrustumMesh,MouseLookXRot,MouseLookYRot,0.0,"xyz")
	      iMeshLocationSet(FrustumMeshShadow,actorshelllocation)
	      iMeshEulerOrientationSet(FrustumMeshShadow,MouseLookXRot,MouseLookYRot,0.0,"xyz")
	   EndIf
	
	   ' Scan the terrain body downward from current actor location, To see If a polygon is there
	   Local p:D3DXVECTOR3=_D3DXVECTOR3(),n:D3DXVECTOR3=_D3DXVECTOR3()
	   If (iBodyScan(Terrain,ActorLocation,_D3DXVECTOR3(0.0,-(SHELL_RADIUS+0.1),0.0),0.05,p,n) >= 0)
	      ' A polygon was detected below the actor
	
	      ' Set current minimal altitude allowed For the actor (rest height).
	      ' The Y coordinate of the scan contact point is used as a reference.
	      Local restheight:Float = p.y+SHELL_RADIUS+0.05
	
	      ' Manage actor vertical position
	
	      If (ActorLocation.y > restheight)
	         ' The actor is currently above the rest height, simply move it downward by
	         ' applying vertical speed To it (add vertical speed To its Y coordinate).
	         ActorLocation.y :+ ActorVerticalSpeed
	      Else
	         ' The actor is currently at Or below the rest height
	
	         ' If [Ctrl] key is pressed move the actor five meters above its current location,
	         ' To simulate a jump. If it is Not pressed simply ensure that the actor is at rest
	         ' height And set vertical speed To zero. 
	         If (iKeyClick(DIK_LCONTROL))
	            ActorLocation.y :+ 5.0
	         Else
	            ActorLocation.y = restheight
	            ActorVerticalSpeed = 0.0
	         EndIf
	      EndIf
	
	      ' Manage actor walking
	
	      ' Build a walking direction vector based on arrow key presses
	      Local walkdirection:D3DXVECTOR3
	      walkdirection=_D3DXVECTOR3(0.0,0.0,0.0)
	      If (iKeyDown(DIK_UP)) walkdirection.PlusEq(_D3DXVECTOR3(0.0,0.0,0.1))
	      If (iKeyDown(DIK_DOWN)) walkdirection.PlusEq(_D3DXVECTOR3(0.0,0.0,-0.1))
	      If (iKeyDown(DIK_LEFT)) walkdirection.PlusEq(_D3DXVECTOR3(-0.05,0.0,0.0))
	      If (iKeyDown(DIK_RIGHT)) walkdirection.PlusEq(_D3DXVECTOR3(0.05,0.0,0.0))
	      ' Store walk direction vector length
	      Local walklength:Float = iVectorLength(walkdirection)
	
	      ' Compute the actual walking direction (walkerorientation)
	      ' If in normal mode it is the current camera orientation.
	      ' If in debug mode, it is the current frustum Object orientation.
	      ' Both are defined by mouse orientation (see above).
	      Local walkerorientation:D3DXQUATERNION=_D3DXQUATERNION()
	      If Not DebugViewMode
			iCameraOrientation(Camera,walkerorientation)
	      Else 
			iMeshOrientation(FrustumMesh,walkerorientation)
		 EndIf
	      ' Transform the arrow-key-defined walk direction vector so
	      ' that it becomes relative To the actual walker orientation.
	      iVectorRotate(walkdirection,walkdirection,walkerorientation)
	      ' cancel any vertical component from it
	      walkdirection.y = 0.0
	      ' set Final vector length To original arrow-key-defined vector length
	      iVectorLengthSet(walkdirection,walkdirection,walklength)
	
	      ' Manage sloping ground effect
	
	      ' Scan again the terrain body downward, along a line 1 meter in front of the actor
	      Local scanorigin:D3DXVECTOR3=_D3DXVECTOR3()
	      iVectorLengthSet(scanorigin,walkdirection,SHELL_RADIUS)
	      scanorigin.PlusEq(ActorLocation)
	      ' Set the scan as Long as 20 meters To make sure steep polygons are also detected
	      scanorigin.y :+ 10.0
	      If (iBodyScan(Terrain,scanorigin,_D3DXVECTOR3(0.0,-20.0,0.0),SHELL_RADIUS,p,n) >= 0)
	         ' A polygon was detected.
	         ' Vector 'n' is now the polygon normal, a unit vector representing
	         ' its absolute orientation. For example, the normal of a perfectly
	         ' horizontal polygon is perfectly upright (0.0,1.0,0.0).
	         ' So, scale the normal To produce a vector proportional To walklength...
	         n.y = 0.0 ' (we ignore vertical component And consider it Null)
	         n.x = iFloatInterpolate(n.x,-0.75,0.75,-1.0,1.0,True) * walklength
	         n.z = iFloatInterpolate(n.z,-0.75,0.75,-1.0,1.0,True) * walklength
	         ' ...And add it To current walk direction vector. Because the normal vector
	         ' is contrary To walking vector For upward sloping ground, And vice-versa,
	         ' this will reduce Or increment walking vector length.
	         walkdirection.PlusEq(n)
	         ' In our example, we don't want the actor to be accelerated by sloping ground.
	         ' So make sure walk direction vector is never longer than computed walklength.
	         If (iVectorLength(walkdirection) > walklength) iVectorLengthSet(walkdirection,walkdirection,walklength)
	      EndIf
	      ' Finally apply the walkdirection vector To current actor location, To move it.
	      ActorLocation.PlusEq(walkdirection)
	   Else
	      ' No polygons were detected below the actor. Just apply vertical speed (falling)
	      ActorLocation.y :+ ActorVerticalSpeed
	   EndIf
	
	   ' Increase vertical speed (downward). The actor falls faster And faster Until
	   ' the terrain is detected And the vertical speed is set To zero (see above).
	   ActorVerticalSpeed :+ -0.002
	
	   ' Manage contact with solid objects
	
	   ' In our example we only check For collisions with the standard ball
	   Local force:D3DXVECTOR3=_D3DXVECTOR3()
	   If (iBodyBodyFeedback(ActorShellBall,-1,0,Null,Null,force,Null,Null,Null))
	      ' A collision was detected. Use the collision force vector To bounce
	      ' the actor back. That is, add a reduced version of the force vector To
	      ' current actor location vector.
	      ActorLocation.PlusEq(force.Mult(0.05))
	   EndIf
	
	   ' Manage the spring
	
	   ' Simply set the target location For the spring To current actor location
	   iSpringTargetBodySet(Spring,Null,ActorLocation)
	
	   ' Manage user controls
	
	   If (iKeyClick(DIK_SPACE))
	      If Not DebugViewMode
	         ' Space key was pressed And we are Not in debug mode: switch To debug mode
	         DebugViewMode = True
	         ' Show actor shell And frustum meshes
	         iMeshShow(ActorShellMesh)
	         iMeshShow(ActorShellMeshShadow)
	         iMeshShow(FrustumMesh)
	         iMeshShow(FrustumMeshShadow)
	         ' Displace a little bit the actor so we can see it from current camera location
	         ActorLocation.z :- 5.0
	      Else
	         ' Space key was pressed And we are in debug mode: switch To 1st-person view mode
	         DebugViewMode = False
	         ' Hide actor shell And frustum meshes
	         iMeshHide(ActorShellMesh)
	         iMeshHide(ActorShellMeshShadow)
	         iMeshHide(FrustumMesh)
	         iMeshHide(FrustumMeshShadow)
	      EndIf
	   EndIf
	
	   ' Manage text messages
	
	   If (DebugViewMode)
	      iPrintCentered("DEBUG VIEW MODE",_D3DXVECTOR2(0.0,11.0),_D3DXVECTOR2(1.0,0.5),0.1,FontWhite,0.0)
	      iPrintCentered("Visualizing actor's shell and viewing frustum",_D3DXVECTOR2(0.0,10.0),_D3DXVECTOR2(0.7,0.5),0.1,FontWhite,0.0)
	      iPrintCentered("USE MOUSE TO LOOK AROUND, [SPACE] TO SWITCH TO NORMAL MODE",_D3DXVECTOR2(0.0,9.0),_D3DXVECTOR2(0.4,0.35),0.1,FontWhite,0.0)
	   Else
	      iPrintCentered("1ST PERSON VIEW MODE",_D3DXVECTOR2(0.0,11.0),_D3DXVECTOR2(1.0,0.5),0.1,FontWhite,0.0)
	      iPrintCentered("What the technique looks like in an actual game",_D3DXVECTOR2(0.0,10.0),_D3DXVECTOR2(0.7,0.5),0.1,FontWhite,0.0)
	      iPrintCentered("USE MOUSE TO LOOK AROUND, [SPACE] TO SWITCH TO DEBUG MODE",_D3DXVECTOR2(0.0,9.0),_D3DXVECTOR2(0.4,0.35),0.1,FontWhite,0.0)
	   EndIf
	
	   ' Manage standard ball gravity
	
	   iBodyAccelerationApply(Ball,_D3DXVECTOR3(0.0,-9.80665,0.0))
	
	   iWindowTitleSet(iFrameRate(True))	
	EndMethod
EndType
